
--[[ Libraries ]]
require('util')
require('config/config')

--[[ Classes ]]
local Tracker = require('lua/tracker')
CSV = require('csv')()
CSV_Coil = require('csv')()

--[[ Initialization ]]
local function UpdateSettings()
    global.tickrate = settings.global['induction-charging-tickrate'].value
    global.score.base = settings.global['induction-charging-score-base'].value
    global.score.neighbour = settings.global['induction-charging-score-neighbour'].value
    global.score.neighbourWithoutWhitespace = settings.global['induction-charging-score-neighbour-without-whitespace'].value
    global.score.minimum = global.score.base
    global.score.maximum = global.score.base
        + (global.score.neighbour * 8)
        + (global.score.neighbourWithoutWhitespace * 8)
end

local function UpdateGlobals()
    global.trackers = global.trackers or {}
    global.score = global.score or {}
    
    UpdateSettings()
end

-- First run
script.on_init(UpdateGlobals)

-- Every run
script.on_load(function()

    -- Fix metatables
    for _, tracker in pairs(global.trackers or {}) do
        -- Fix tracker which in turns fixes all its children
        classifyInstance(tracker, Tracker)
        tracker:Fix()
    end
    
end)

-- Configs changed
script.on_configuration_changed(UpdateGlobals)

--[[ Utilities ]]
function wtf(...)
    local args = {...}
    for k,v in pairs(args) do
        if type(v) ~= 'string' or v == nil then
            args[k] = stringify(v)
        end
    end
    for k,v in pairs(game.players) do
        v.print(table.concat(args, ', '))
    end
end

local function getPlayer(event)
    return game.players[event.player_index]
end

local function getTracker(event)
    if not global.trackers[event.player_index] then
        global.trackers[event.player_index] = Tracker(getPlayer(event))
    end
    return global.trackers[event.player_index]
end

local function deleteTracker(event)
    local tracker = global.trackers[event.player_index]
    if tracker and tracker:Exists() then
        tracker:Destroy()
    end

    global.trackers[event.player_index] = nil
end

--[[ Events ]]

-- Game updates
script.on_event(defines.events.on_tick, function(event)

    -- Update all shadows
    if event.tick % global.tickrate == 0 then
        for id, tracker in pairs(global.trackers) do
            tracker:Tick()
        end
    end

end)

-- Player equips/unequips armor
script.on_event(defines.events.on_player_armor_inventory_changed, function(event)
    local tracker = getTracker(event)
    tracker:UpdateGrid()
end)

-- Player adds equipment to equipment grid
script.on_event(defines.events.on_player_placed_equipment, function(event)
    local tracker = getTracker(event)
    tracker:UpdateGrid()
end)

-- Player removes equipment from equipment grid
script.on_event(defines.events.on_player_removed_equipment, function(event)
    local tracker = getTracker(event)
    tracker:UpdateGrid()
end)

-- Player joins
script.on_event(defines.events.on_player_joined_game, function(event)
    local tracker = getTracker(event)
    tracker:UpdateGrid()
end)

-- Player leaves
script.on_event(defines.events.on_player_left_game, function(event)
    local tracker = getTracker(event)
    tracker:UpdateGrid()
    deleteTracker(event)
end)

-- Player dies
script.on_event(defines.events.on_player_died, function(event)
    local ply = getPlayer(event)
    local tracker = getTracker(event)

    tracker:UpdateGrid()
end)

-- Research finishes
script.on_event(defines.events.on_research_finished, function(event)

    local research = event.research
    if table.contains(config.technologies, nil, research.name) then

        -- Recalculate throughput and efficiency for all grids
        for _, tracker in pairs(global.trackers) do
            if tracker.grid then
                tracker.grid:UpdateCalculation()
            end
        end

    end
    
end)

-- Player changes Surface
script.on_event(defines.events.on_player_changed_surface, function(event)
    -- Honestly? We don't care. The entity will respawn if the surface has changed.
    -- No extra action is necessary here; it'll only impede our progress in the
    -- tick update of the trackers.
end)

-- Chat for debugging
script.on_event(defines.events.on_console_chat, function(event)
    CSV:Write('output')
    CSV_Coil:Write('coil')
end)

script.on_event(defines.events.on_runtime_mod_setting_changed, UpdateSettings)