
local Equipment = require('equipment')

local g = {}

function g:New(tracker, grid)
    self.tracker = tracker
    self.grid = grid
    self.width = grid and grid.width or 0
    self.height = grid and grid.height or 0

    self.batteries = nil
    self.coils = nil
    self.map = nil
    
    self.throughput = 0
    self.efficiency = 0

    self.energy = 0
    self.capacity = 0
    self.requested = 0

    self:Update()
end

function g:Fix()
    if self.map then
        -- Set map metatable(s)
        remapN(self.map, 2)

        -- Fix all items
        for x, column in pairs(self.map) do
            for y, item in pairs(column) do

                -- Check if it has already been fixed.
                -- This happens when there are pieces of equipment
                -- larger than 1x1 and therefore are stored in the map
                -- multiple times!
                if not item.Fix then
                    classifyInstance(item, Equipment)
                    item:Fix()
                end

            end
        end
    end
end

function g:Set(grid)
    self.grid = grid
    if grid then
        self:Update()
    end
end

function g:Distribute(received)
    CSV:Value('received', received)
    CSV:Value('capacity', self.capacity)
    CSV:Value('requested', (self.requested or 0))
    CSV:Value('receive%', received / (self.requested or 0) * 100)
    CSV:Value('throughput', self.throughput)
    CSV:Value('efficiency%', self.efficiency * 100)

    -- If we're not valid, return zero
    if not self:Exists() then
        return 0
    end

    -- Check how much power all our batteries have and put it in self.energy
    self:UpdateEnergy()

    -- Calculate how much energy we've received and how much
    -- we have in total. We also account for efficiency here.
    local realReceived = received * self.efficiency
    local totalStored = math.ceil(self.energy + realReceived)

    -- Determine how many joules we can provide in [global.tickrate] ticks (joules/tickrate)
    local providableJoules = self.throughput * (global.tickrate / 60)

    -- Check how much space there is left for us to fill (joules)
    local spaceAvailable = math.max(0, self.capacity - totalStored)

    -- Fill as much as possible within our throughput (joules)
    local maxFillableSpace = math.min(spaceAvailable, providableJoules)

    -- Calculate how much we should request to negate losses by efficiency (joules)
    local toRequest = math.ceil(maxFillableSpace / self.efficiency)
    self.requested = toRequest

    -- Now that we've determined how much power we will request in the next
    -- [global.tickrate] ticks, we can distribute the power we've received.

    -- Do nothing if we received nothing
    if received <= 0 then
        return toRequest
    end

    -- Distribute power evenly
    local fraction = math.min(1, totalStored / self.capacity) -- cap at 1
    for _, battery in pairs(self.batteries) do

        -- Round the 'gainz' down in order to prevent "creating" energy by rounding errors
        battery.equipment.energy = math.floor(fraction * battery.equipment.max_energy)

    end

    -- Request whatever amount we need for the next [global.tickrate] ticks.
    return toRequest
end

function g:UpdateEnergy()
    local energy = 0
    for _, battery in pairs(self.batteries) do
        energy = energy + battery:Energy()
    end
    self.energy = energy -- table operations are slow (premature optimization, hehe)
    return energy
end

function g:Update()
    self:UpdateEquipment()
    self:UpdateCalculation()
end

function g:UpdateEquipment()
    -- Reset internal equipment list/map
    self.batteries = {}
    self.coils = {}
    self.map = mapN(2)
    self.capacity = 0

    -- Check if we're valid
    if not self:Exists() then
        return
    end

    -- Iterate over new equipment
    for _, eq in pairs(self.grid.equipment) do

        -- Itemize it
        local item = Equipment(self.grid, eq)

        -- Store battery if applicable
        if item.isBattery then
            table.insert(self.batteries, item)
            self.capacity = self.capacity + item:MaxEnergy()
        end

        -- Store coil if applicable
        if item.isCoil then
            table.insert(self.coils, item)
        end

        -- Store in map
        for x = item.x1, item.x2, 1 do
            for y = item.y1, item.y2, 1 do
                self.map[x][y] = item
            end
        end
    end

    -- Update the neighbour counts
    for _, coil in pairs(self.coils) do
        coil:CheckNeighbours(self.map)
    end

    -- Determine amount of neighbours with whitespace.
    -- We do this after CheckNeighbours. The alternative was to
    -- do it recursively in CheckNeighbours but I chose not to
    -- do so due to infinite looping concerns.
    for _, coil in pairs(self.coils) do
        coil:CheckBorders()
    end
end

function g:UpdateCalculation()

    if not self:Exists() then
        return
    end

    -- Update our efficiency and throughput
    self.efficiency = self.tracker:GetEfficiency()    
    self.throughput = 0
    local baseThroughput = self.tracker:GetThroughput()

    -- Add throughput per coil based on multiplier
    for _, coil in pairs(self.coils) do        
        self.throughput = self.throughput +
            math.floor(baseThroughput * coil:CalculateMultiplier())
    end
end

function g:Exists()
    return self.grid ~= nil and self.grid.valid
end

return classify(g)