
local Grid = require('grid')
local Gui = require('gui')
local Shadow = require('shadow')

local t = {}

function t:New(ply)
    self.entity = ply
    self.grid = Grid(self, nil)
    self.shadow = Shadow(self)
    self.gui = Gui(self)

    self.lastThroughput = 0
end

function t:Fix()
    -- If we already have a Grid instance we will set its metatable.
    -- Otherwise, we create a new instance.
    if self.grid then
        classifyInstance(self.grid, Grid)
        self.grid:Fix()
    else
        self.grid = Grid(self, nil)
    end

    -- We'd like to do the same for the Shadow, but this is a little
    -- more complicated. In some save files during testing the self.shadow
    -- field was set to the actual entity. We also account fot that here
    -- by checking if it has a 'valid' field.
    if self.shadow then
        if self.shadow.valid then
            -- It's an entity.. whoops
            self.shadow = Shadow(self)
        else
            classifyInstance(self.shadow, Shadow)
        end
    else
        self.shadow = Shadow(self)
    end

    -- Fix gui instance; it is always present.
    classifyInstance(self.gui, Gui)
    self.gui:Fix()
end

function t:Tick()

    CSV:Value('tick', game.tick)
    CSV:Value('player', self.entity.name)

    -- Amount of energy our shadow received
    local energy = self.shadow:Energy()
    local needGui = false

    if self.shadow:Exists() then
        CSV:Value('1_energy', tostring(self.shadow.entity.energy))
        CSV:Value('1_buffer', tostring(self.shadow.entity.electric_buffer_size))
        CSV:Value('1_flow', tostring(self.shadow.entity.electric_input_flow_limit))
    else
        CSV:Value('1_energy', '')
        CSV:Value('1_buffer', '')
        CSV:Value('1_flow', '')
    end
    
    -- Scale the received energy back to watts
    self.lastThroughput = (energy / global.tickrate) * 60 *
        (self.grid:Exists() and self.grid.efficiency or 0)

    -- Distribute over grid, provided that it exists
    if self.grid:Exists() then

        -- Show the GUI if the grid contains coils
        needGui = needGui or table.count(self.grid.coils) > 0
    end
        
    -- Distribute the received energy to the grid,
    -- which in turn tells us how much more we need to request.
    -- Returns zero if grid is invalid.
    local requested = self.grid:Distribute(energy)

    -- Move shadow entity to correct position if it exists
    -- and request the amount of power we need, if any
    self.shadow:Move()
    self.shadow:Request(requested)

    -- Update the GUI
    self.gui:Visible(needGui)
    self.gui:Update()

    CSV:Next()
end

function t:UpdateGrid()

    -- Get armor inventory if any
    local inv = self.entity.get_inventory(defines.inventory.player_armor)
    if inv == nil then
        self.grid:Set(nil)
        return
    end

    -- Get the grid for said armor
    local armor = inv[1]
    if armor and armor.valid and armor.valid_for_read and armor.grid ~= nil then
        self.grid:Set(armor.grid)
    else
        self.grid:Set(nil)
    end
end

function t:GetTier()
    local highest = 0
    local techs = self.entity.force.technologies

    for num, tier in pairs(config.tiers) do
        if num > 0 then
            if techs["induction" .. tostring(num)].researched then
                highest = num
            end
        end
    end
    return highest
end

function t:GetEfficiency()
    return config.tiers[self:GetTier()].efficiency
end

function t:GetThroughput()
    return config.tiers[self:GetTier()].throughput
end

function t:Alive()
    return self.entity and self.entity.valid
end

return classify(t)